package scenegraph;

import java.util.*;

import javax.vecmath.Matrix4d;
import javax.vecmath.Vector3d;

public class Node implements ReferenceFrame{
	Container parent;
	public Container getParent(){
		return parent;
	}
	public void setParent(Container new_parent){
		
		if (new_parent == parent){
			return;
		}
		
		if (new_parent != parent){
			if (parent != null)
				parent.children.remove(this);
		}
		
		parent = new_parent;

		if (new_parent != null){
			parent.children.add(this);
		}

	}
	
	
	
	
	
	
	public Matrix4d getAbsoluteTransformation(){
		ArrayList<Matrix4d> chain = new ArrayList<Matrix4d>();
		Node current = this;
		if (current instanceof Transformable)
			chain.add(((Transformable)current).matrix);
		while(true){
			current = current.parent;
			if (current == null || current instanceof Scene)
				break;
			if (current instanceof Transformable)
				chain.add(((Transformable)current).matrix);
		}
		Matrix4d absolute_matrix = new Matrix4d();
		absolute_matrix.setIdentity();
		for(Matrix4d m : chain){
			absolute_matrix.mul(m, absolute_matrix);
		}
		return absolute_matrix;
		
	}

	public Vector3 transformTo(Vector3d localPt, ReferenceFrame reference){
		Vector3 p = new Vector3();
		if (reference == null){
			p.set(localPt);
			return p;
		}
		
		Matrix4d local_to_absolute = this.getAbsoluteTransformation();
		Matrix4d reference_to_absolute = reference.getAbsoluteTransformation();
		reference_to_absolute.invert();
		Matrix4d local_to_reference = new Matrix4d();
		local_to_reference.mul(reference_to_absolute, local_to_absolute);
		
		//p.set(localPt);
		//local_to_reference.transform(p);
		p = transform(local_to_reference, localPt);
		
		return p;
	}
	Vector3 transform(Matrix4d m, Vector3d v){
		return new Vector3( 
				m.m00*v.x+m.m01*v.y+m.m02*v.z+m.m03, 
				m.m10*v.x+m.m11*v.y+m.m12*v.z+m.m13, 
				m.m20*v.x+m.m21*v.y+m.m22*v.z+m.m23);
	}
}
