package scenegraph;

import java.awt.Graphics;

import javax.vecmath.Matrix4d;

import javax.media.opengl.*;
import javax.media.opengl.glu.*;



import com.sun.opengl.util.*;
import java.nio.*;
import java.util.*;
import java.util.HashMap;

public class Renderer implements GLEventListener {
	public Renderer(){
	}
	RenderTarget renderTarget;
	public RenderTarget createRenderTarget() {
		renderTarget = new RenderTarget();	//GLCanvas
		renderTarget.renderer = this;
		renderTarget.addGLEventListener(this);
		
//		final Animator animator = new Animator(renderTarget);
//		animator.start();

		return renderTarget;
	}
	public OffscreenRenderTarget createOffscreenRenderTarget(){
		offscreenRenderTarget = new OffscreenRenderTarget();
		offscreenRenderTarget.renderer = this;
		return offscreenRenderTarget;
	}
	OffscreenRenderTarget offscreenRenderTarget;
	
	void display_scenegraph(GL gl, GLU glu){
		Camera camera = renderTarget.camera;
	
		Node current = camera;
		while(true){
			current = current.parent;
			if (current instanceof Scene || current == null){
				break;
			}
			if (current instanceof Transformable){
				Transformable transformable = (Transformable) current;
				
				Matrix4d m = new Matrix4d();
				m.set(transformable.matrix);
				//Debug.print(m);
				m.invert();
				
				gl.glMultMatrixd(getOpenGLMatrix(m), 0);
			}
		}
		
		float[] bc = ((Scene)current).background.color;
		gl.glClearColor(bc[0],bc[1],bc[2],bc[3]);
		gl.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT);

		
		// light
		set_light(gl, glu, current);
		
		//System.out.println("display_scenegraph "+this+" "+current);

		// visual
		display_scenegraph(gl, glu, current);
		
	}
	double[] getOpenGLMatrix(Matrix4d matrix){
		return new double[]{
				matrix.m00,
				matrix.m10,
				matrix.m20,
				matrix.m30,
				matrix.m01,
				matrix.m11,
				matrix.m21,
				matrix.m31,
				matrix.m02,
				matrix.m12,
				matrix.m22,
				matrix.m32,
				matrix.m03,
				matrix.m13,
				matrix.m23,
				matrix.m33	
		};
	}
	void set_light(GL gl, GLU glu, Node current){
		gl.glPushMatrix();
		
		if (current instanceof Transformable){
			Transformable transformable = (Transformable) current;
			gl.glMultMatrixd(transformable.getOpenGLMatrix(), 0);
		}

		if (current instanceof Container){
			java.util.ArrayList<Node> children = (java.util.ArrayList<Node>)  (((Container)current).children).clone();
			for(Node child : children){
				set_light(gl, glu, child);	
			}
		}
		
		if (current instanceof AmbientLight){
			float[] glfLightAmbient = ((AmbientLight)current).color;
			gl.glLightfv(GL.GL_LIGHT1, GL.GL_AMBIENT, glfLightAmbient, 0);
			gl.glEnable(GL.GL_LIGHT1);
		}
		if (current instanceof DirectionalLight){
			float[] glfLightDiffuse = ((DirectionalLight)current).color;
			gl.glLightfv(GL.GL_LIGHT0, GL.GL_DIFFUSE, glfLightDiffuse, 0);
			float[] xyz = {0,0,-1, 0};
			gl.glLightfv(GL.GL_LIGHT0, GL.GL_POSITION, xyz, 0);
			gl.glEnable(GL.GL_LIGHT0);
		}
		
		gl.glPopMatrix();
	}
	
	
	void display_scenegraph(GL gl, GLU glu, Node current){
		gl.glPushMatrix();
		
		if (current instanceof Transformable){
			Transformable transformable = (Transformable) current;
			gl.glMultMatrixd(transformable.getOpenGLMatrix(), 0);
		}
		
		//synchronized (current) 
		if (current instanceof Container){
			java.util.ArrayList<Node> children = (java.util.ArrayList<Node>)  (((Container)current).children).clone();
			for(Node child : children){
			//for(Node child : ((Container)current).children){
				display_scenegraph(gl, glu, child);	
			}
		}
		
		if (current instanceof Visual){
			display_visual(gl, glu, (Visual) current);
		}
		gl.glPopMatrix();
	}
	//float[] color = {1,1,1,0};
	
	void display_visual(GL gl, GLU glu, Visual visual){
		gl.glLoadName(get_index(visual));

		if (update_requested(visual)) {
			if (visual.display_list_id == -1)
				visual.display_list_id = gl.glGenLists(1);
			gl.glNewList(visual.display_list_id, GL.GL_COMPILE);

			display_visual_oneside(gl, glu, visual, visual.frontFacingAppearance, true);
			display_visual_oneside(gl, glu, visual, visual.backFacingAppearance, false);
			
			gl.glEndList();
			
			//clear_update_requests(visual);
			
		}
		gl.glCallList(visual.display_list_id);
		
		
		//gl.glPopName();
		
	}
	
	HashMap<Object, Integer> compiled_version = new HashMap();
	int get_comiled_version(Object object){
		Integer integer = compiled_version.get(object);
		if (integer == null)
			return -1;
		return integer.intValue();
	}
	boolean update_requested(Visual visual){
		
		if (compiled_version.get(visual) != null || visual.version != get_comiled_version(visual)){
			compiled_version.put(visual,  visual.version);
			return true;
		}
		if (visual.frontFacingAppearance != null && 
				visual.frontFacingAppearance.version != get_comiled_version(visual.frontFacingAppearance)){
			compiled_version.put(visual.frontFacingAppearance,  visual.frontFacingAppearance.version);
			return true;
		}
		if (visual.backFacingAppearance != null && 
				visual.backFacingAppearance.version != get_comiled_version(visual.backFacingAppearance)){
			compiled_version.put(visual.backFacingAppearance,  visual.backFacingAppearance.version);
			return true;
		}
		if (visual.geometry != null && visual.geometry.version != get_comiled_version(visual.geometry)){
			compiled_version.put(visual.geometry,  visual.geometry.version);
			return true;
		}
		return false;
	}
	
//	boolean update_requested(Visual visual){
//		if (visual.update_requested)
//			return true;
//		if (visual.frontFacingAppearance != null && visual.frontFacingAppearance.update_requested)
//			return true;
//		if (visual.backFacingAppearance != null && visual.backFacingAppearance.update_requested)
//			return true;
//		if (visual.geometry instanceof IndexedTriangleArray && ((IndexedTriangleArray)visual.geometry).update_requested )
//			return true;
//		if (visual.geometry instanceof LineArray && ((LineArray)visual.geometry).update_requested )
//			return true;
//		return false;
//	}
//	void clear_update_requests(Visual visual){
//		visual.update_requested = false;
//		if (visual.frontFacingAppearance != null)
//			visual.frontFacingAppearance.update_requested = false;
//		if (visual.backFacingAppearance != null)
//			visual.backFacingAppearance.update_requested = false;
//		if (visual.geometry instanceof IndexedTriangleArray)
//			((IndexedTriangleArray)visual.geometry).update_requested = false;
//		if (visual.geometry instanceof LineArray)
//			((LineArray)visual.geometry).update_requested = false;
//	}
	void display_visual_oneside(GL gl, GLU glu, Visual visual, Appearance appearance, boolean front){
		if (appearance == null)
			return;
		if (appearance.shadingStyle == ShadingStyle.SMOOTH)
			gl.glShadeModel(GL.GL_SMOOTH);
		else if (appearance.shadingStyle == ShadingStyle.FLAT)
			gl.glShadeModel(GL.GL_FLAT);
		
		gl.glMaterialfv(GL.GL_FRONT, GL.GL_DIFFUSE, appearance.diffuseColor, 0);
		gl.glMaterialfv(GL.GL_FRONT, GL.GL_AMBIENT, appearance.ambientColor, 0);
		
		
		if (visual.backFacingAppearance != null){
			gl.glMaterialfv(GL.GL_BACK, GL.GL_DIFFUSE, visual.backFacingAppearance.diffuseColor, 0);
		}
		
		// enable texture
		if (appearance.diffuseColorMap != null){
			gl.glEnable(gl.GL_TEXTURE_2D);
			gl.glTexEnvf(gl.GL_TEXTURE_ENV, gl.GL_TEXTURE_ENV_MODE, gl.GL_DECAL);
			int textureID = appearance.diffuseColorMap.textureID;
			if (textureID == -1)
				textureID = appearance.diffuseColorMap.textureID 
						  = Texture.create_texture(gl, (java.awt.image.BufferedImage) appearance.diffuseColorMap.image);
			gl.glBindTexture(gl.GL_TEXTURE_2D, textureID);
		}
		else {
			gl.glDisable(gl.GL_TEXTURE_2D);
		}
		
		if (visual.geometry instanceof IndexedTriangleArray)
			displayIndexedTriangleArray(gl, visual, front);
		else if (visual.geometry instanceof LineArray)
			displayLineArray(gl, visual);

		

	}
	void displayLineArray(GL gl,Visual visual){
		gl.glColor3d(0,0,0);
		gl.glBegin(GL.GL_LINES);
		for (Vertex3d v : ((LineArray)visual.geometry).vertices) {
			gl.glVertex3d(v.position.x, v.position.y, v.position.z);
		}
		gl.glEnd();
	}

	void displayIndexedTriangleArray(GL gl, Visual visual, boolean front){
		if (visual.frontFacingAppearance.fillingStyle == FillingStyle.WIREFRAME){
			gl.glColor3d(0,0,0);
			gl.glBegin(GL.GL_LINES);
			for (int i = 0; i < ((IndexedTriangleArray)visual.geometry).indices.length/3; i++) {
				for(int j=0; j<3; j++){
					Vertex3d v = ((IndexedTriangleArray)visual.geometry).vertices[((IndexedTriangleArray)visual.geometry).indices[i*3+j]];
					gl.glVertex3d(v.position.x, v.position.y, v.position.z);
					v = ((IndexedTriangleArray)visual.geometry).vertices[((IndexedTriangleArray)visual.geometry).indices[i*3+ (j+1)%3]];
					gl.glVertex3d(v.position.x, v.position.y, v.position.z);
				}
			}
			gl.glEnd();
		}
		else if (visual.frontFacingAppearance.fillingStyle == FillingStyle.SOLID){
			gl.glBegin(GL.GL_TRIANGLES);
			for (int i = 0; i < ((IndexedTriangleArray)visual.geometry).indices.length/3; i++) {
				for(int j=0; j<3; j++){
					int index = front ? i*3+2-j : i*3+j;
					Vertex3d v = ((IndexedTriangleArray)visual.geometry).vertices[((IndexedTriangleArray)visual.geometry).indices[index ]];
					gl.glTexCoord2f(v.textureCoordinate0.x, v.textureCoordinate0.y);
					gl.glNormal3d(v.normal.x,v.normal.y,v.normal.z);
					gl.glVertex3d(v.position.x, v.position.y, v.position.z);
				}
			}
		}
	    gl.glEnd();
	}
	
	
	
	
	
	
	
	//
	// paint overlay
	//
//	public void paint(Graphics g) {
//		renderTarget.paint(g);
//		renderTarget.renderedListener.rendered(g);	// paint Overlay
//	}

	public void repaint(){
		renderTarget.repaint();
	}

	
	public void init(GLAutoDrawable drawable) {
		GL gl = drawable.getGL();
//		gl.setSwapInterval(1);

//		gl.glEnable(GL.GL_CULL_FACE);
//		gl.glEnable(GL.GL_LIGHTING);
//		gl.glEnable(GL.GL_DEPTH_TEST);
//		gl.glEnable(GL.GL_NORMALIZE);

//		renderTarget.repaint();
	}

	float aspect_ratio = 1;
	public void reshape(GLAutoDrawable drawable, int x, int y, int width, int height) {
		GL gl = drawable.getGL();

		System.err.println("GL_VENDOR: " + gl.glGetString(GL.GL_VENDOR));
		System.err.println("GL_RENDERER: " + gl.glGetString(GL.GL_RENDERER));
		System.err.println("GL_VERSION: " + gl.glGetString(GL.GL_VERSION));

		aspect_ratio =  (float) width / (float) height;

		//System.err.println("aspect ratio: " + aspect_ratio);

	}

	public void displayChanged(GLAutoDrawable drawable, boolean modeChanged,
			boolean deviceChanged) {
	}

	GLAutoDrawable drawable;
	
	public float d = 1f;
	public void __display(GLAutoDrawable drawable) {
	    GL gl = drawable.getGL();
		gl.glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
		gl.glClear(GL.GL_COLOR_BUFFER_BIT);
		gl.glColor3f(1.0f, 1.0f, 1.0f);
		gl.glOrtho(-1.0, 1.0, -1.0, 1.0, -1.0, 1.0);
		gl.glBegin(GL.GL_POLYGON);
		gl.glVertex2f(-0.5f*d, -0.5f*d);
		gl.glVertex2f(-0.5f*d, 0.5f*d);
		gl.glVertex2f(0.5f*d, 0.5f*d);
		gl.glVertex2f(0.5f*d, -0.5f*d);
		gl.glEnd();
		gl.glFlush();
	}

	
	public void display(GLAutoDrawable drawable) {
		if (renderTarget.renderedListener != null){
			Graphics g = renderTarget.getGraphics();
			renderTarget.renderedListener.rendered(g);
			g.dispose();
		}
		

		this.drawable = drawable;
		
		drawable.getContext().makeCurrent();

		
		GL gl = drawable.getGL();
		GLU glu = new GLU();
		
		gl.glEnable(GL.GL_CULL_FACE);
		gl.glEnable(GL.GL_LIGHTING);
		gl.glEnable(GL.GL_DEPTH_TEST);
		gl.glEnable(GL.GL_NORMALIZE);

		
		display_main(gl, glu, false);
		
		stored_gl = gl;
		stored_glu = glu;

		
		drawable.getContext().release();
		
		

	}
	
	public void display_main(GL gl, GLU glu, boolean selection_mode){
		
		gl.glPushMatrix();
		
		gl.glMatrixMode(GL.GL_PROJECTION);
		gl.glLoadIdentity();
		
		if (selection_mode){
		    int viewport[] = new int[4];
		    gl.glGetIntegerv(GL.GL_VIEWPORT, viewport, 0);
		    glu.gluPickMatrix((double) pick_x,
		        (double) (viewport[3] - pick_y), 
		        1.0, 1.0, viewport, 0);
		}		
		
		if (renderTarget.camera instanceof OrthographicCamera){
			OrthographicCamera c = (OrthographicCamera) renderTarget.camera;
			if (Float.isNaN(c.x0))
				gl.glOrtho(c.y0*aspect_ratio, c.y1*aspect_ratio, c.y0, c.y1, c.near, c.far);
			else if (Float.isNaN(c.y0))
				gl.glOrtho(c.x0, c.x1, c.x0/aspect_ratio, c.x1/aspect_ratio, c.near, c.far);
			else
				gl.glOrtho(c.x0, c.x1, c.y0, c.y1, c.near, c.far);
		}
		else { 
			PerspectiveCamera c = (PerspectiveCamera) renderTarget.camera;
			glu.gluPerspective((float)(180*c.verticalViewingAngle/Math.PI), aspect_ratio, c.nearClippingPlaneDistance, c.farClippingPlaneDistance);
		}
		
		gl.glMatrixMode(GL.GL_MODELVIEW);
		gl.glLoadIdentity();
		
		gl.glScalef(1,1,-1);				// directX <-> openGL			

		// store view transform
		if (!selection_mode){
			gl.glGetDoublev(GL.GL_MODELVIEW_MATRIX, modelview, 0);
			gl.glGetDoublev(GL.GL_PROJECTION_MATRIX, projection, 0);
			gl.glGetIntegerv(GL.GL_VIEWPORT, viewport,0);
		}
		
		
 		//
		//	screen update
		//
		display_scenegraph(gl, glu);

		gl.glPopMatrix();
		
		gl.glFlush();

		//drawable.getContext().release();
	}
	
	
	
	
	
	
	int BUFSIZE = 256;
	int pick_x;
	int pick_y;
	PickInfo pick(int x, int y){
		
		set_index(get_scene(renderTarget.camera), 0);
		
		int[] selectBuf = new int[BUFSIZE];
	    IntBuffer selectBuffer = BufferUtil.newIntBuffer(BUFSIZE);

	    drawable.getContext().makeCurrent();
	    
	    
	    pick_x = x;
	    pick_y = y;
	    
	    stored_gl.glSelectBuffer(BUFSIZE, selectBuffer);
	    stored_gl.glRenderMode(GL.GL_SELECT);

	    stored_gl.glInitNames();
	    stored_gl.glPushName(0);
	    
	    display_main(stored_gl, stored_glu, true);
	
	    int hits = stored_gl.glRenderMode(GL.GL_RENDER);
	    if (hits == 0)
	    	return null;
	    
	    selectBuffer.get(selectBuf);
	    PickInfo pickInfo = processHits(hits, selectBuf);
		
	    
	    drawable.getContext().release();
	    
	    
	    return pickInfo;
	}
	Scene get_scene(Camera camera){
		Node node = camera;
		while(true){
			if (node instanceof Scene)
				return (Scene) node;
			if (node == null)
				return null;
			node = node.parent;
		}
	}
	java.util.ArrayList<Visual> index_to_visual = new java.util.ArrayList<Visual>();
	int get_index(Visual visual){
		for(int i=0; i<index_to_visual.size(); i++){
			if (index_to_visual.get(i) == visual)
				return i;
		}
		return -1;
	}
	int set_index(Node node, int index){
		if (node instanceof Visual){
			index_to_visual.add((Visual)node);
			index++;
		}
		if (node instanceof Container){
			for(Node child : ((Container)node).children){
				index = set_index(child, index);
			}
		}
		return index;
	}

	PickInfo processHits(int hits, int buffer[])
	  {
	    int names, ptr = 0;

	    VisualDepth[] visualdepths = new VisualDepth[hits];
	    
	    // System.out.println("hits = " + hits);
	    // ptr = (GLuint *) buffer;
	    for (int i = 0; i < hits; i++)
	    { /* for each hit */
	    	Visual visual = null;
	    	
	      names = buffer[ptr];
	      //System.out.println(" number of names for hit = " + names);
	      ptr++;
	      //System.out.println("  z1 is " + buffer[ptr]);
	      int depth = buffer[ptr];
	      ptr++;
	      //System.out.println(" z2 is " + buffer[ptr]);
	      ptr++;
	      //System.out.print("\n   the name is ");
	      for (int j = 0; j < names; j++)
	      { /* for each name */
	        //System.out.println("" + buffer[ptr]);
	    	visual = index_to_visual.get(buffer[ptr]);
	        ptr++;
	      }
	      
	      visualdepths[i] = new VisualDepth(visual, depth);
	    }

	    Arrays.sort(visualdepths, new java.util.Comparator(){
	    	public int compare(Object a, Object b){
	    		int d0 = ((VisualDepth)a).depth;
	    		int d1 = ((VisualDepth)b).depth;
	    		if (d0 < d1)
	    			return -1;
	    		if (d0 > d1)
	    			return 1;
	    		return 0;
	    	}
	    });
	    
	    PickInfo pickInfo = new PickInfo();
	    pickInfo.visuals = new Visual[hits];

	   // System.out.println("ProcessHits sorted");
	    for(int i=0; i<hits; i++){
	    	pickInfo.visuals[i] = visualdepths[i].visual;
	    	// System.out.println(" "+visualdepths[i].visual + " "+ visualdepths[i].depth);
	    }
	    return pickInfo;
	   

	  }
	class VisualDepth {
		Visual visual;
		int depth;
		VisualDepth(Visual _visual, int _depth){
			visual = _visual;
			depth = _depth;
		}
	}
	
	
	
	
	
	
	
	
	
	
	//
	// view transform
	//
	GL stored_gl;
	GLU stored_glu;
	double[] modelview = new double[16];//fr[s擾
	double[] projection = new double[16];//es擾
	int[] viewport = new int[4];//r[|[g擾
	public double[] world_coords_to_screen_coords(double x, double y, double z){
		double[] v = new double[3];
		stored_glu.gluProject(x,y,z, modelview, 0, projection, 0, viewport, 0, v, 0);
		v[1] = renderTarget.getHeight()-v[1];
		return v;
	}
	public double[] screen_coords_to_world_coords(double x, double y, double z){
		y = renderTarget.getHeight()-y;
		double[] v = new double[3];
		stored_glu.gluUnProject(x, y, z, modelview, 0, projection, 0, viewport, 0, v,0);
		return v;
	}
}
