package scenegraph;

import javax.media.opengl.*;
import java.awt.*;
import java.awt.color.ColorSpace;
import java.awt.image.*;
import java.io.*;
import java.nio.*;
import java.util.*;
import java.awt.geom.AffineTransform;

import javax.imageio.ImageIO;

public class Texture {
	
	public static int create_texture(GL gl, BufferedImage image){
		
		int[] textureID = {-1};//new int [1];
	
		TextureLoader textureLoader = new TextureLoader(gl);
		Texture texture = textureLoader.getTexture(image, true);
		
		if(textureLoader.isOK()) {
			gl.glPixelStorei(gl.GL_UNPACK_ALIGNMENT, 1);

			gl.glGenTextures(1, textureID, 0);
			gl.glBindTexture(gl.GL_TEXTURE_2D, textureID[0]);
			
			gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_WRAP_S, gl.GL_REPEAT);
			gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_WRAP_T, gl.GL_REPEAT);
			gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MAG_FILTER, gl.GL_NEAREST);
			gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MIN_FILTER, gl.GL_NEAREST);

			gl.glTexImage2D(gl.GL_TEXTURE_2D, 0, texture.getGLFormat(),
							texture.getImageWidth(), texture.getImageHeight(), 0,
							texture.getGLFormat(), gl.GL_UNSIGNED_BYTE,
							texture.getTextureBuffer());
		}
		return textureID[0];
	}
	
	static int load_texture(GL gl, String filename){
		
		int[] textureID = new int [1];
	
		TextureLoader textureLoader = new TextureLoader(gl);
		Texture texture = textureLoader.getTexture(filename, true);
		
		if(textureLoader.isOK()) {
			gl.glPixelStorei(gl.GL_UNPACK_ALIGNMENT, 1);

			gl.glGenTextures(1, textureID, 0);
			gl.glBindTexture(gl.GL_TEXTURE_2D, textureID[0]);
			
			gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_WRAP_S, gl.GL_REPEAT);
			gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_WRAP_T, gl.GL_REPEAT);
			gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MAG_FILTER, gl.GL_NEAREST);
			gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MIN_FILTER, gl.GL_NEAREST);

			gl.glTexImage2D(gl.GL_TEXTURE_2D, 0, texture.getGLFormat(),
							texture.getImageWidth(), texture.getImageHeight(), 0,
							texture.getGLFormat(), gl.GL_UNSIGNED_BYTE,
							texture.getTextureBuffer());
		}
		return textureID[0];
	}	
	
	
	
	
	
	
	
	
	
	
	
	

//-----------------------------------------------------------------------------
// DATA MEMBERS

	private String resourceName;

	/**
	 * height of the image loaded from disk
	 */
	private int imageHeight;

	/**
	 * width of the image loaded from disk
	 */
	private int imageWidth;

	/**
	 * height of the texture. This may differ from the image height
	 * if the image loaded from disk did not have a height that was
	 * a power of two. This texture height could potentially be set to
	 * the nearest power of two greater than or equal to the image height.
	 * This essentially allows us to scale up an image loaded from disk so
	 * that it meets the OpenGL requirement that it have a height that is a
	 * power of two. These considerations should be handled by the TextureLoader.
	 */
	private int texHeight; // texture height

	/**
	 * width of the texture. This may differ from the image wisth
	 * if the image loaded from disk did not have a width that was
	 * a power of two. This texture width could potentially be set to
	 * the nearest power of two greater than or equal to the image width.
	 * This essentially allows us to scale up an image loaded from disk so
	 * that it meets the OpenGL requirement that it have a width that is a
	 * power of two. These considerations should be handled by the TextureLoader.
	 */
	private int texWidth;

	/**
	 * the ratio of the image width to the texture width
	 */
	private float widthRatio;

	/**
	 * the ratio of the image height to the texture height
	 */
	private float heightRatio;

	/**
	 * the type of image this texture was loaded from. Either JPG or TGA.
	 */
	private int imageType;

	/**
	 * used as srcPixelFormat when building mipmaps
	 */
	private int glFormat;

	/**
	 * byte buffer containing pixel data
	 */
	private ByteBuffer buffer;

	static Texture lastbind = null;

	/**
	 * constant defined to represent an image type of .jpg
	 */
	public final static int JPG=0;

	/**
	 * constant defined to represent an image type of .tga
	 */
	public final static int TGA=1;
	public final static int GIF=2;

//-----------------------------------------------------------------------------
// CONSTRUCTORS

	/**
	 * This constructor sets the texture type (JPG or TGA) and stores the
	 * resource path.
	 *
	 * @param resourceName	filename of texture file
	 */
	public Texture(String resourceName) {
		if(resourceName.endsWith(".tga")) {
			//System.out.println("Texture is TGA");
			imageType = TGA;
		}
		else if(resourceName.endsWith(".jpg")){
			//System.out.println("Texture is JPG");
			imageType = JPG;
		}
		else if(resourceName.endsWith(".gif")){
			//System.out.println("Texture is JPG");
			imageType = GIF;
		}
		else
			System.out.println("Texture: Unknown image type");
		this.resourceName = resourceName;
	} // end constructor
	
	public Texture(){
	}
	

//-----------------------------------------------------------------------------
// METHODS

	/**
	 * This method sets the texture's byte buffer.
	 *
	 * @param buffer	the byte buffer to set for the texture
	 */
	public void setTextureBuffer(ByteBuffer  buffer) {
		this.buffer = buffer;
	} // end setTextureBuffer

	/**
	 * This method returns the texture's byte buffer.
	 *
	 * @return	ByteBuffer - the texture's byte buffer
	 */
	public ByteBuffer getTextureBuffer() {
		return buffer;
	} // end getTextureBuffer

	/**
	 * This method returns the texture's GL Format.
	 *
	 * @return	int - the texture's GL Format
	 */
	public int getGLFormat(){
		return glFormat;
	} // end getGLFormat

	/**
	 * This method sets the texture's GL Format.
	 *
	 * @param newFormat	the GL Format to set for the texture
	 */
	public void setGLFormat(int newFormat){
		glFormat = newFormat;
	} // end setGLFormat

	/**
	 * This method sets the texture's image height.
	 *
	 * @param height	the image height to set for the texture
	 */
	public void setImageHeight(int imgHeight) {
		this.imageHeight = imgHeight;
	} // end setImageHeight

	/**
	 * This method sets the texture's image width.
	 *
	 * @param width	the image width to set for the texture
	 */
	public void setImageWidth(int imgWidth) {
		this.imageWidth = imgWidth;
	} // end setImageWidth

	/**
	 * This method returns the texture's image height.
	 *
	 * @return	int - the texture's image height
	 */
	public int getImageHeight() {
		return imageHeight;
	} // end getImageHeight

	/**
	 * This method returns the texture's image width.
	 *
	 * @return	int - the texture's image width
	 */
	public int getImageWidth() {
		return imageWidth;
	} // end getImageWidth

	/**
	 * This method returns the texture's height ratio.
	 * That is, the ratio of the images's height to the
	 * texture's height.
	 *
	 * @return	float - the texture's height ratio
	 */
	public float getHeightRatio() {
		return heightRatio;
	} // end getHeightRatio

	/**
	 * This method returns the texture's width ratio.
	 * That is, the ratio of the image's width to the
	 * texture's width.
	 *
	 * @return	float - the texture's width ratio
	 */
	public float getWidthRatio() {
		return widthRatio;
	} // end getWidthRatio

	/**
	 * This method returns the texture's image type.
	 *
	 * @return	int - the texture's image type, either Texture.JPG or Texture.TGA
	 */
	public int getImageType() {
		return imageType;
	} // end getImageType

	/**
	 * This method sets the texture's height and updates the
	 * heightRatio.
	 *
	 * @param texHeight	the height to set for the texture
	 */
	public void setTextureHeight(int texHeight) {
		this.texHeight = texHeight;
		setHeightRatio();
	} // end setTextureHeight

	/**
	 * This method sets the texture's width and updates the
	 * widthRatio.
	 *
	 * @param texWidth	the width to set for the texture
	 */
	public void setTextureWidth(int texWidth) {
		this.texWidth = texWidth;
		setWidthRatio();
	} // end setTextureWidth

	/**
	 * This method sets the texture's height ratio to the ratio of
	 * the image's height to the texture's height.
	 */
	private void setHeightRatio() {
		if (texHeight != 0) {
			heightRatio = ((float) imageHeight)/texHeight;
		} // end if
	} // end setHeightRatio

	/**
	 * This method sets the texture's width ratio to the ratio of
	 * the image's width to the texture's width.
	 */
	private void setWidthRatio() {
		if (texWidth != 0) {
			widthRatio = ((float) imageWidth)/texWidth;
		} // end if
	} // end setWidthRatio

	/**
	 * This method method returns the file suffix (extension) of the provided filename.
	 *
	 * @param filePath	the filename to determine the file suzffix for
	 * @return	String - the file's suffix (extension)
	 */
	private String fileSuffix(String filePath) {
		int dotPos = filePath.lastIndexOf(".");
		return filePath.substring(dotPos+1);
	}//end fileFromPath method

} // end class Texture
















/**
 * A utility class to load textures for JOGL.
 *
 * @author Kevin Glass modified by Bill Clark and Ryan Mauer
 * @version 22-Apr-2004
 */
class TextureLoader
{

//-----------------------------------------------------------------------------
// DATA MEMBERS
	/**<code>gl, glu</code> references to the caller's gl and glu objects
	 * to allow OpenGL calls
	 */
	 GL gl;
	// GLU glu;

	/**
	 * <code>true</code> if an error occured during the last call to
	 * <code>getTexture()</code> or <code>false</code> otherwise
	 */
	private boolean error = false;

	/**
	 * ColorModel allowing alpha color component
	 */
	private ColorModel glAlphaColorModel;

	/**
	 * ColorModel disallowing alpha color component
	 */
	private ColorModel glColorModel;

//-----------------------------------------------------------------------------
// CONSTRUCTORS

	/**
	 * This constructor creates the two default ColorModels.
	 */
	public TextureLoader(GL gl) {	//, GLU glu
		this.gl = gl;
		//this.glu = glu;

		glAlphaColorModel = new ComponentColorModel(ColorSpace.getInstance(ColorSpace.CS_sRGB),
						new int[] {8,8,8,8},
						true,
						false,
						ComponentColorModel.TRANSLUCENT,
						DataBuffer.TYPE_BYTE);

		glColorModel = new ComponentColorModel(ColorSpace.getInstance(ColorSpace.CS_sRGB),
						new int[] {8,8,8,0},
						false,
						false,
						ComponentColorModel.OPAQUE,
						DataBuffer.TYPE_BYTE);
	} // end constructor

//-----------------------------------------------------------------------------
// METHODS

	/**
	 * This method loads a Texture from the specified filename and
	 * returns it. If an error occured while loading the texture, null is
	 * returned and <code>isOK()</code> returns false. Otherwise, <code>isOK()</code>
	 * returns true, indicating that texture loading succeeded.
	 *
	 * @param resourceName		the filename of the texture to load
	 * @param forcePowersOfTwo	<code>true</code> to force the height and width
	 *									of the texture to be powers of two or
	 *									<code>false</code> otherwise
	 * @return	Texture - the loaded Texture object
	 */
	public Texture getTexture(String resourceName, boolean forcePowersOfTwo)
	{
		Texture tex = null;

		//System.out.println("Loading texture "+resourceName);
		Texture texture = new Texture(resourceName);
		BufferedImage bufferedImage = null;
		try {
			bufferedImage = loadImage(resourceName);
		}
		catch( Exception e )
		{
			System.out.println("Unable to load texture "+ e );
			error = true;
			return null;
		}
		error = false;
		if(texture.getImageType() == Texture.JPG) {
			// Flip Image
			AffineTransform tx = AffineTransform.getScaleInstance(1, -1);
			tx.translate(0, -bufferedImage.getHeight(null));
			AffineTransformOp op = new AffineTransformOp(tx,
					 AffineTransformOp.TYPE_NEAREST_NEIGHBOR);
			bufferedImage = op.filter(bufferedImage, null);
		}

		// Getting the real Width/Height of the Texture in the Memory
		int realWidth, realHeight;
		if(forcePowersOfTwo) {
			// force the width and height of the created texture to be powers of two
			realWidth = get2Fold(bufferedImage.getWidth());
			realHeight = get2Fold(bufferedImage.getHeight());
		} // end if
		else {
			// DO NOT force the width and height of the created texture to be powers of two
			realWidth = bufferedImage.getWidth();
			realHeight = bufferedImage.getHeight();
		} // end else

		// set the height and width of image loaded from disk
		texture.setImageWidth(realWidth);
		texture.setImageHeight(realHeight);

		if (bufferedImage.getColorModel().hasAlpha()) {
			texture.setGLFormat(GL.GL_RGBA);
		}
		else {
			texture.setGLFormat(GL.GL_RGB);
		}

		// convert that image into a byte buffer of texture data
		ByteBuffer textureBuffer = convertImageData(bufferedImage,texture,forcePowersOfTwo);
		texture.setTextureBuffer(textureBuffer);
		return texture;

	} // end getTexture
	
	
	public Texture getTexture(BufferedImage bufferedImage, boolean forcePowersOfTwo)
	{
		Texture tex = null;

		Texture texture = new Texture();

		
		// Getting the real Width/Height of the Texture in the Memory
		int realWidth, realHeight;
		if(forcePowersOfTwo) {
			// force the width and height of the created texture to be powers of two
			realWidth = get2Fold(bufferedImage.getWidth());
			realHeight = get2Fold(bufferedImage.getHeight());
		} // end if
		else {
			// DO NOT force the width and height of the created texture to be powers of two
			realWidth = bufferedImage.getWidth();
			realHeight = bufferedImage.getHeight();
		} // end else

		// set the height and width of image loaded from disk
		texture.setImageWidth(realWidth);
		texture.setImageHeight(realHeight);

		if (bufferedImage.getColorModel().hasAlpha()) {
			texture.setGLFormat(GL.GL_RGBA);
		}
		else {
			texture.setGLFormat(GL.GL_RGB);
		}

		// convert that image into a byte buffer of texture data
		ByteBuffer textureBuffer = convertImageData(bufferedImage,texture,forcePowersOfTwo);
		texture.setTextureBuffer(textureBuffer);
		return texture;

	} // end getTexture
	
	

	/**
	 * This method should be called after each call to <code>getTexture()</code>.
	 * It returns <code>false</code> if an error occured during texture loading or
	 * <code>true</code> if texture loading succeeded.
	 *
	 * @return	<code>true</code> if texture loading succeeded or <code>false</code> otherwise
	 */
	public boolean isOK()
	{
		return !error;

	} // end isOK

	/**
	 * This method returns the nearest power of two greater than or equal
	 * to the specified value.
	 *
	 * @param fold	the value to find power of two above
	 * @return	int - the nearest power of two greater than or equal to the specified value
	 */
	private int get2Fold(int fold)
	{
		int ret = 2;
		while (ret < fold) {
			ret *= 2;
		}
		return ret;
	} // end get2Fold

	/**
	 * This method returns a ByteBuffer representation of the specified
	 * BufferedImage and Texture data.
	 * <p>
	 * If <code>forcePowersOfTwo</code> is <code>true</code> then this method performs
	 * steps necessary to force the width and height of the returned texture
	 * ByteBuffer to be powers of two. It does this by creating a new
	 * BufferedImage with a width and height that are the nearest power
	 * of two greater than or equal to the parameter's. The parameter's pixel
	 * data is then copied into the new BufferedImage. Thus, no scaling is
	 * done but any area outside the original texture size will be filled with
	 * a solid background color. The background color depends on what is used
	 * by default when creating a new BufferedImage.
	 * <p>
	 * If <code>forcePowersOfTwo</code> is <code>false</code> then this method uses the
	 * width and height of the BufferedImage parameter without modifying them.
	 *
	 * @param bufferedImage		the BufferedImage to use for conversion
	 * @param texture				the Texture to use for conversion
	 * @param forcePowersOfTwo	<code>true</code> to force the height and width
	 *									of the texture to be powers of two or
	 *									<code>false</code> otherwise
	 * @return	ByteBuffer - the ByteBuffer containing the texture data
	 */
	private ByteBuffer convertImageData(BufferedImage bufferedImage, Texture texture, boolean forcePowersOfTwo)
	{
		ByteBuffer imageBuffer = null;
		WritableRaster raster;
		BufferedImage texImage;

		// Determine the texture height and width.
		int texWidth, texHeight;
		if(forcePowersOfTwo) {
			// We require that the texture height and width be powers of two
			// by using get2fold()
			texWidth = get2Fold(bufferedImage.getWidth());
			texHeight = get2Fold(bufferedImage.getHeight());
		} // end if
		else {
			// DO NOT require that the texture height and width be powers of two
			texWidth = bufferedImage.getWidth();
			texHeight = bufferedImage.getHeight();
		} // end else

		// Set the texture width and height.
		// This also updates the width ratio and height ratio where:
		// height ratio = ((float) image height) / (texture height)
		// width ratio = ((float) image width) / (texture width)
		texture.setTextureHeight(texHeight);
		texture.setTextureWidth(texWidth);

		String [] propNames = bufferedImage.getPropertyNames();
		//System.out.println(bufferedImage);

		// create a Raster
		// create a BufferedImage using the raster
		if (bufferedImage.getColorModel().hasAlpha()) {
			raster = Raster.createInterleavedRaster(DataBuffer.TYPE_BYTE,texWidth,texHeight,4,null);
			texImage = new BufferedImage(glAlphaColorModel,raster,false,new Hashtable());
		} // end if
		else {
			raster = Raster.createInterleavedRaster(DataBuffer.TYPE_BYTE,texWidth,texHeight,3,null);
			texImage = new BufferedImage(glColorModel,raster,false,new Hashtable());
		} // end else

		// if (forcePowersOfTwo == true) then:
		// - this basically copies the BufferedImage parameter into a new BufferedImage
		//   that is guaranteed to have a width and height that are a power of two.
		// - if the BufferedImage parameter already had a width and height that were
		//   powers of two, then there isn't any difference between it and the new BufferedImage.
		// - up to this point, we basically just forced the width and height to be powers
		//   of two if they were not already.
		Graphics g = texImage.getGraphics();
//		g.drawImage(bufferedImage,0,0, texWidth, texHeight, 0,0,
//					bufferedImage.getWidth(), bufferedImage.getHeight(), 
//					null);
		g.drawImage(bufferedImage,0,texHeight, texWidth, 0, 0,0,		// ㉺]II
					bufferedImage.getWidth(), bufferedImage.getHeight(), 
					null);

		// get the byte data
		byte[] data = ((DataBufferByte) texImage.getRaster().getDataBuffer()).getData();
		return ByteBuffer.wrap(data);
/*
		// allocate space for and store the pixel data
		imageBuffer = ByteBuffer.allocateDirect(data.length);
		imageBuffer.order(ByteOrder.nativeOrder());
		imageBuffer.put(data, 0, data.length);

		// return the pixel data
		return imageBuffer;
*/
	} // end convertImageData

	/**
	 * This method loads a BufferedImage from the specified file using ImageIO.
	 *
	 * @param ref	the filename of the image to load
	 * @return	BufferedImage - the loaded BufferedImage
	 * @exception	IOException - thrown if the indicated file cannot be found
	 * @exception	Exception - thrown if the BufferedImage is null after loading
	 */
	public BufferedImage loadImage(String ref) throws Exception
	{
		File file = new File(ref);

		if (!file.isFile()) {
			throw new IOException("Cannot find: "+ref);
		}
		ImageIO.setUseCache(false);
		BufferedImage bufferedImage =
				ImageIO.read(new BufferedInputStream(new FileInputStream(file)));

		if(bufferedImage == null)
			throw new Exception("BufferedImage null after loadImage()");

		return bufferedImage;

	} // end loadImage

} // end class TextureLoader

