package scenegraph;

import java.util.ArrayList;

import javax.vecmath.*;
public class Transformable extends Container {
	Matrix4d matrix;
	public Transformable(){
		matrix = new Matrix4d();
		matrix.setIdentity();
	}
	
	double[] getOpenGLMatrix() {
		return new double[]{
			matrix.m00,
			matrix.m10,
			matrix.m20,
			matrix.m30,
			matrix.m01,
			matrix.m11,
			matrix.m21,
			matrix.m31,
			matrix.m02,
			matrix.m12,
			matrix.m22,
			matrix.m32,
			matrix.m03,
			matrix.m13,
			matrix.m23,
			matrix.m33

		};
	}
	public void setPosition(Vector3d p){
		matrix.m03 = p.x;
		matrix.m13 = p.y;
		matrix.m23 = p.z;
		
//		Matrix4d d = new Matrix4d();
//		d.set(p);
//		
//		matrix.mul(d, matrix);
	}
	public void rotate(Vector3d axis, double angle){
		AxisAngle4d a = new AxisAngle4d(axis.x, axis.y, axis.z, angle);
		Matrix4d m = new Matrix4d();
		m.set(a);
		matrix.mul(matrix, m);
		//matrix.mul(m, matrix);// wrong
	}
	public void translate(Vector3d v){
		matrix.m03 += v.x;
		matrix.m13 += v.y;
		matrix.m23 += v.z;
	}	
	public void setPosition(Vector3d p, ReferenceFrame reference){
		if (reference == null){
			setPosition(p);
			return;
		}
		if (parent == null){
			setPosition(p);
			return;
		}

		Matrix4d local_to_absolute = parent.getAbsoluteTransformation();
		Matrix4d reference_to_absolute = reference.getAbsoluteTransformation();
		reference_to_absolute.invert();
		Matrix4d local_to_reference = new Matrix4d();
		local_to_reference.mul(reference_to_absolute, local_to_absolute);

		matrix.mul(local_to_reference, matrix);
		
		matrix.m03 = p.x;
		matrix.m13 = p.y;
		matrix.m23 = p.z;
		
		local_to_reference.invert();
		
		matrix.mul(local_to_reference, matrix);
		
	}
	public void translate(Vector3d v, ReferenceFrame reference){
		if (reference == null){
			translate(v);
			return;
		}


		Matrix4d local_to_absolute = parent.getAbsoluteTransformation();
		Matrix4d reference_to_absolute = reference.getAbsoluteTransformation();
		reference_to_absolute.invert();
		Matrix4d local_to_reference = new Matrix4d();
		local_to_reference.mul(reference_to_absolute, local_to_absolute);

		matrix.mul(local_to_reference, matrix);
		
		matrix.m03 += v.x;
		matrix.m13 += v.y;
		matrix.m23 += v.z;
		
		local_to_reference.invert();
		
		matrix.mul(local_to_reference, matrix);
		
	}
	
	
	// TODO
	public void scale(Vector3d v, ReferenceFrame reference){
		if (reference == null){
			matrix.m00 *= v.x;
			matrix.m11 *= v.y;
			matrix.m22 *= v.z;
			return;
		}

		if (parent == null)
			return;
		
		Matrix4d local_to_absolute = parent.getAbsoluteTransformation();
		Matrix4d reference_to_absolute = reference.getAbsoluteTransformation();
		reference_to_absolute.invert();
		Matrix4d local_to_reference = new Matrix4d();
		local_to_reference.mul(reference_to_absolute, local_to_absolute);

		matrix.mul(local_to_reference, matrix);
		
		matrix.m00 *= v.x;
		matrix.m11 *= v.y;
		matrix.m22 *= v.z;
		
		local_to_reference.invert();
		
		matrix.mul(local_to_reference, matrix);
		
	}
	public void setScale(Vector3d v, ReferenceFrame reference){
		if (reference == null || parent == null){
			matrix.m00 = v.x;
			matrix.m11 = v.y;
			matrix.m22 = v.z;
			return;
		}

		Matrix4d local_to_absolute = parent.getAbsoluteTransformation();
		Matrix4d reference_to_absolute = reference.getAbsoluteTransformation();
		reference_to_absolute.invert();
		Matrix4d local_to_reference = new Matrix4d();
		local_to_reference.mul(reference_to_absolute, local_to_absolute);

		matrix.mul(local_to_reference, matrix);
		
		matrix.m00 = v.x;
		matrix.m11 = v.y;
		matrix.m22 = v.z;
		
		local_to_reference.invert();
		
		matrix.mul(local_to_reference, matrix);
		
	}
	
	public void rotate(Vector3d axis, double angle, ReferenceFrame reference){
		if (reference == null){
			rotate(axis, angle);
			return;
		}
		AxisAngle4d a = new AxisAngle4d(axis.x, axis.y, axis.z, angle);
		Matrix4d am = new Matrix4d();
		am.set(a);

		Matrix4d local_to_absolute = parent.getAbsoluteTransformation();
		Matrix4d reference_to_absolute = reference.getAbsoluteTransformation();
		reference_to_absolute.invert();
		Matrix4d local_to_reference = new Matrix4d();
		local_to_reference.mul(reference_to_absolute, local_to_absolute);

		am.mul(am, local_to_reference);
		
		local_to_reference.invert();
		
		am.mul(local_to_reference, am);
		
		matrix.mul(am, matrix);
		
		
	}
	
	public void setAbsoluteTransformation(Matrix4d new_absolute_matrix){
		Matrix4d absolute_matrix_parent = parent.getAbsoluteTransformation();
		absolute_matrix_parent.invert();
		
		matrix.mul(new_absolute_matrix, absolute_matrix_parent);
	}
	public Matrix3d getAxes(ReferenceFrame reference){
		Matrix4d _matrix = new Matrix4d();
		if (reference == null){
			_matrix = matrix;
		}
		else {
			Matrix4d local_to_absolute = parent.getAbsoluteTransformation();
			Matrix4d reference_to_absolute = reference.getAbsoluteTransformation();
			reference_to_absolute.invert();
			Matrix4d local_to_reference = new Matrix4d();
			local_to_reference.mul(reference_to_absolute, local_to_absolute);
		
			_matrix.mul(local_to_reference, matrix);
		}
		
		
		Matrix3d m =  new Matrix3d();
		m.m00 = _matrix.m00;
		m.m01 = _matrix.m01;
		m.m02 = _matrix.m02;
		m.m10 = _matrix.m10;
		m.m11 = _matrix.m11;
		m.m12 = _matrix.m12;
		m.m20 = _matrix.m20;
		m.m21 = _matrix.m21;
		m.m22 = _matrix.m22;
		return m;
	}
	public Vector3 getPosition(ReferenceFrame reference){
		//Vector3d p = new Vector3d(matrix.m03, matrix.m13, matrix.m23);
		Vector3d p = new Vector3d();
		
		return transformTo(p, reference);
	}
	
	public Vector3d[] getOrientation(ReferenceFrame reference){
		Matrix4d _matrix = new Matrix4d();
		if (reference == null){
			_matrix = matrix;
		}
		else {
			Matrix4d local_to_absolute = parent.getAbsoluteTransformation();
			Matrix4d reference_to_absolute = reference.getAbsoluteTransformation();
			reference_to_absolute.invert();
			Matrix4d local_to_reference = new Matrix4d();
			local_to_reference.mul(reference_to_absolute, local_to_absolute);
		
			_matrix.mul(local_to_reference, matrix);
		}
		
		Vector3d x = new Vector3d(_matrix.m00, _matrix.m10, _matrix.m20);
		Vector3d y = new Vector3d(_matrix.m01, _matrix.m11, _matrix.m21);
		Vector3d z = new Vector3d(_matrix.m02, _matrix.m12, _matrix.m22);
		
		return new Vector3d[]{z, x, y};
	}
	
	static void setOrientation(Matrix4d _matrix, Vector3d x, Vector3d y, Vector3d z){
		_matrix.m00 = x.x;
		_matrix.m10 = x.y;
		_matrix.m20 = x.z;
		_matrix.m01 = y.x;
		_matrix.m11 = y.y;
		_matrix.m21 = y.z;
		_matrix.m02 = z.x;
		_matrix.m12 = z.y;
		_matrix.m22 = z.z;
		
	}
	public void setTransformation(Matrix4d m, ReferenceFrame reference){
		if (reference == null){
			matrix = m;
			return;
		}
		Matrix4d local_to_absolute = parent.getAbsoluteTransformation();
		Matrix4d reference_to_absolute = reference.getAbsoluteTransformation();
		reference_to_absolute.invert();
		Matrix4d local_to_reference = new Matrix4d();
		local_to_reference.mul(reference_to_absolute, local_to_absolute);

		
		local_to_reference.invert();

		matrix = m;
		
		matrix.mul(local_to_reference, matrix);		
	}
	public void setLocalTransformation(Matrix4d m){
		matrix = m;
	}
	public Matrix4d getTransformation(ReferenceFrame reference){
		if (reference == null)
			return matrix;
		
		Matrix4d local_to_absolute = parent.getAbsoluteTransformation();
		Matrix4d reference_to_absolute = reference.getAbsoluteTransformation();
		reference_to_absolute.invert();
		Matrix4d local_to_reference = new Matrix4d();
		local_to_reference.mul(reference_to_absolute, local_to_absolute);
		
		Matrix4d m = new Matrix4d();
		
		m.mul(local_to_reference, matrix);
		
		return m;
	}
	public void setOrientation(Vector3d z, Vector3d y, ReferenceFrame reference){
		z = new Vector3d(z);
		z.normalize();
		y = new Vector3d(y);
		y.normalize();
		
		Vector3d x = new Vector3d();
		x.cross(y, z);
		x.normalize();

		y.cross(z, x);
		
		if (reference == null){
			setOrientation(matrix, x, y, z);
			return;
		}
		
		Matrix4d local_to_absolute = parent.getAbsoluteTransformation();
		Matrix4d reference_to_absolute = reference.getAbsoluteTransformation();
		reference_to_absolute.invert();
		Matrix4d local_to_reference = new Matrix4d();
		local_to_reference.mul(reference_to_absolute, local_to_absolute);

		matrix.mul(local_to_reference, matrix);
		
		setOrientation(matrix, x, y, z);

		local_to_reference.invert();
		
		matrix.mul(local_to_reference, matrix);
		
	}
	public void pointAt(ReferenceFrame target, Vector3d offset, Vector3d up_vec, ReferenceFrame reference ){
		Matrix4d target_to_reference = getRelativeTransformation(target, reference);
		Vector3d target_p = new Vector3d();
		//target_to_reference.transform(target_p);
		target_p = transform(target_to_reference, target_p);
		target_p.add(offset);
		
		Matrix4d local_to_reference = getRelativeTransformation(this, reference);
		Vector3d local_p = new Vector3d();
		//local_to_reference.transform(local_p);
		local_p = transform(local_to_reference, local_p);
		
		Vector3d z = new Vector3d();
		z.sub(target_p, local_p);

		setOrientation(z, up_vec, reference);
	}
	
	public Matrix4d getRelativeTransformation(ReferenceFrame local, ReferenceFrame reference){
		Matrix4d local_to_absolute = local.getAbsoluteTransformation();
		Matrix4d reference_to_absolute = reference.getAbsoluteTransformation();
		reference_to_absolute.invert();
		Matrix4d local_to_reference = new Matrix4d();
		local_to_reference.mul(reference_to_absolute, local_to_absolute);
		return local_to_reference;
	}

	
}
